import os
import shlex
import subprocess
import unittest


def run_cmd(cmd):
    """Run Commands"""
    process = subprocess.Popen(shlex.split(cmd), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    (stdout, stderr) = process.communicate()
    return stdout, stderr, process.wait()


class Test(unittest.TestCase):
    def setUp(self):
        os.environ['APP_NAME'] = __name__
        os.environ['TEMPLATE_FOLDER'] = os.path.join(os.path.dirname(__file__), 'public/templates')
        os.environ['STATIC_FILES'] = os.path.join(os.path.dirname(__file__), 'public/static')

    def tearDown(self):
        os.environ.clear()

    def test_hash(self):
        from bast.hash import Hash
        password = 'basttestpassword'
        hashed = Hash.encrypt(password)
        return hashed

    def test_compare(self):
        from bast.hash import Hash
        password = "basrtest".encode('utf-8')
        hashed = Hash.encrypt(password)

        print(hashed, str(password))
        self.assertEqual(Hash.compare(password, hashed), True)

    def test_script_server(self):
        from bast.view import script
        script_link = script("http://bootstrap.js")
        self.assertEqual('<script type="text/javascript" src="http://bootstrap.js"></script>', script_link, "Not equal")

    def test_css_server(self):
        from bast.view import css
        css_link = css('http://mycssfile.com')
        self.assertEqual('<link rel="stylesheet" href="http://mycssfile.com">', css_link)

    def test_css_local(self):
        from bast.view import css
        css_link = css('my_css_file.css')
        self.assertEqual('<link rel="stylesheet" href="/css/my_css_file.css">', css_link)

    def test_script_local(self):
        from bast.view import script
        script_link = script("my_script.js")
        self.assertEqual('<script type="text/javascript" src="/script/my_script.js"></script>', script_link, "Not equal")



"""
    Bast Web Framework
    (c) Majiyagbe Oluwole <oluwole564@gmail.com>

    For full copyright and license information, view the LICENSE distributed with the Source Code
"""

import os
import subprocess
import sys

from orator.database_manager import DatabaseManager
from orator.migrations import MigrationCreator, Migrator, DatabaseMigrationRepository


class CreateMigration(MigrationCreator):
    """
    Handles the creation of the Migration file. Makes use of the Orator ORM for Migration
    """
    def __init__(self):
        self.path = os.path.abspath('.') + '/database/migrations/'
        if not os.path.exists(self.path):
            os.makedirs(self.path)

    def create_file(self, name, table, create=True):
        return self.create(name, self.path, table=table, create=create)


class Migration(Migrator):
    """
    Handles the actions  to be performed on the Migration files
    """
    def __init__(self):
        """
        Initialize the Orator Migrator Class. Check if the migration table has been created. If not, Create the
        Repository
        """
        check, config = self.get_config()
        if not check:
            print("Error Occurred")
        else:
            self.manager = DatabaseManager(config=config)
            self.path = os.path.abspath('.') + "/database/migrations/"
            self.repository = DatabaseMigrationRepository(resolver=self.manager, table='migrations')

            if not self.repository.repository_exists():
                self.repository.create_repository()

            super().__init__(self.repository, self.manager)

    def run_(self, pretend):
        """
        Run the migration file
        :param pretend: Determines whether to run the migration as a Simulation or not. Defaults to False
        :return:
        """
        self.run(self.path, pretend=pretend)

    def rollback_(self, pretend):
        """
        Roll Back the Last Migration
        :param pretend: Determines whether to run the migration as a Simulation or not. Defaults to False
        :return: int
        """
        return self.rollback(self.path, pretend)

    def reset_(self, pretend):
        """
        Reset all the migrations that have been done
        :param pretend: Determines whether to run the migration as a Simulation or not. Defaults to False
        :return: int
        """
        return self.reset(self.path, pretend)

    @staticmethod
    def get_config():
        """
        Gets the config from the os.environ. This is used to create the config dict for use by the ORM
        :return: str, dict
        """
        db_type = os.environ['DB_TYPE']
        db_host = os.environ['DB_HOST']
        db_user = os.environ['DB_USER']
        db_database = os.environ['DB_NAME']
        db_password = os.environ['DB_PASSWORD']
        db_prefix = os.environ['DB_PREFIX']

        check = Migration.check_packages(db_type)

        return check, {
            db_type: {
                'driver': db_type.strip(),
                'host': db_host.strip(),
                'database': db_database.strip(),
                'user': db_user.strip(),
                'password': db_password.strip(),
                'prefix': db_prefix.strip()
            }
        }

    @staticmethod
    def check_packages(db_name):
        """
        Check if the driver for the user defined host is available. If it is not available, download it using PIP
        :param db_name:
        :return:
        """
        print('Checking for required Database Driver')

        reqs = subprocess.check_output([sys.executable, '-m', 'pip', 'freeze'])
        installed_packages = [r.decode().split('==')[0] for r in reqs.split()]

        # print(installed_packages)

        if db_name.lower() == 'mysql':
            if 'PyMySQL' not in installed_packages:
                print('Installing required Database Driver')
                os.system('pip install pymysql')

        if db_name.lower() == 'postgresql':
            if 'psycopg2-binary' not in installed_packages:
                print('Installing required Database Driver')
                os.system('pip install psycopg2-binary')

        return True



"""
    Bast Web Framework
    (c) Majiyagbe Oluwole <oluwole564@gmail.com>

    For full copyright and license information, view the LICENSE distributed with the Source Code
"""

from .bast import Bast
from .controller import Controller
from .hash import Hash
from .migration import CreateMigration
from .migration import Migration
from .model import Models
from .route import Route
from .session import FileSession
from .session import MemorySession
from .environment import load_env

__version__ = "1.0"




"""
    Bast Web Framework
    (c) Majiyagbe Oluwole <oluwole564@gmail.com>

    For full copyright and license information, view the LICENSE distributed with the Source Code
"""

import os
from jinja2 import Environment, FileSystemLoader, TemplateNotFound, select_autoescape


def script(js_file):
    """
    Returns a route to the JS folder. Renders the JS file in the js folder
    Error 404 is seen in the server log if not found
    :param js_file:
    :return:
    """
    if 'http' in js_file:
        return '<script type="text/javascript" src="' + js_file + '"></script>'
    return '<script type="text/javascript" src="/script/' + js_file + '"></script>'


def css(css_file):
    """
    Returns a route to the CSS File and renders it to the server
    Error 404 is seen in the server log if not found
    :param css_file:
    :return:
    """
    if 'http' in css_file:
        return '<link rel="stylesheet" href="' + css_file + '">'
    return '<link rel="stylesheet" href="/css/' + css_file + '">'


def image(image_file, alt_name="image"):
    """
    Returns a route to the image file and renders it to the server
    Error 404 is thrown in server log if not found
    :param alt_name:
    :param image_file:
    :return:
    """
    if 'http' in image_file:
        return '<img src="' + image_file + '" alt="' + alt_name + '">'
    return '<img src="/images/' + image_file + '" alt="' + alt_name + '">'


# def session():
#     return


class TemplateRendering:
    """
    Base class to load the template directory from the OS Environment TEMPLATE_FOLDER variable
    """

    def __init__(self):
        self.dictionary = {}

    def add_(self, key, _dict_):
        self.dictionary[key] = _dict_
        return self

    def render_template(self, template_name, **kwargs):
        template_dir = os.environ['TEMPLATE_FOLDER']

        env = Environment(loader=FileSystemLoader(template_dir))

        env.globals['css'] = css
        env.globals['script'] = script

        self.dictionary.update(**kwargs)

        try:
            template = env.get_template(template_name)
        except TemplateNotFound:
            raise TemplateNotFound(template_name)
        content = template.render(self.dictionary)
        return content

    @classmethod
    def render_exception(cls, **kwargs):
        template_dir = os.path.dirname(os.path.realpath(__file__)) + "/exception"
        env = Environment(loader=FileSystemLoader(template_dir))

        try:
            template = env.get_template('exception.html')
        except TemplateNotFound:
            raise TemplateNotFound('exception.html')
        content = template.render(**kwargs)
        return content



"""
    Bast Web Framework
    (c) Majiyagbe Oluwole <oluwole564@gmail.com>

    For full copyright and license information, view the LICENSE distributed with the Source Code
"""

import os
import subprocess
import sys

from orator import DatabaseManager
from orator.orm import Model
from colorama import init, Fore


class Models(Model):
    def __init__(self, **attributes):
        super().__init__(**attributes)
        init()
        db = DatabaseManager(self.get_config())
        self.set_connection_resolver(db)

    @staticmethod
    def get_config():
        db_type     = os.getenv('DB_TYPE')
        db_host     = os.getenv('DB_HOST')
        db_user     = os.getenv('DB_USER')
        db_database = os.getenv('DB_NAME')
        db_password = os.getenv('DB_PASSWORD')
        db_prefix   = os.getenv('DB_PREFIX')

        Models.check_packages(db_type)

        return {
            db_type: {
                'driver': db_type.strip(),
                'host': db_host.strip(),
                'database': db_database.strip(),
                'user': db_user.strip(),
                'password': db_password.strip(),
                'prefix': db_prefix.strip()
            }
        }

    @staticmethod
    def check_packages(db_name):
        print(Fore.GREEN + 'Checking for required Database Driver')

        reqs = subprocess.check_output([sys.executable, '-m', 'pip', 'freeze'])
        installed_packages = [r.decode().split('==')[0] for r in reqs.split()]

        if db_name.lower() == 'mysql':
            if 'PyMySQL' not in installed_packages:
                print(Fore.GREEN + 'Installing required Database Driver')
                subprocess.call(['pip', 'install', 'pymysql'])

        if db_name.lower() == 'postgresql':
            if 'psycopg2' not in installed_packages:
                print(Fore.GREEN + 'Installing required Database Driver')
                subprocess.call(['pip', 'install', 'psycopg2'])



"""
    Bast Web Framework
    (c) Majiyagbe Oluwole <oluwole564@gmail.com>

    For full copyright and license information, view the LICENSE distributed with the Source Code
"""

import bcrypt


class Hash:
    """
     Hash Plain Values using Python bcrypt (https://pypi.org/project/bcrypt/)
    Example Usage
    hashed_password = Hash.encrypt(text_to_encrypt)

    To make use of the new Key Defined Function (KDF), use
    hashed_password = Hash.encrypt(text_to_encrypt, method='kdf')

    To check if hashed password is the same as a plain text

    is_same = Hash.compare(plain_text, hashed_text)

    """
    @classmethod
    def encrypt(cls, plain_text, method='normal'):
        plain_text = str(plain_text).encode('utf-8')
        if method is 'kdf':
            return bcrypt.kdf(password=plain_text, salt=cls.gensalt(12), desired_key_bytes=32, rounds=100)
        return bcrypt.hashpw(plain_text, cls.gensalt(12))

    @classmethod
    def compare(cls, plain_text, hash_text):
        plain_text = str(plain_text).encode('utf-8')
        return bcrypt.checkpw(plain_text, hash_text)

    @classmethod
    def gensalt(cls, rounds=12, prefix=b"2b"):
        return bcrypt.gensalt(rounds, prefix)




"""
    Bast Web Framework
    (c) Majiyagbe Oluwole <oluwole564@gmail.com>

    For full copyright and license information, view the LICENSE distributed with the Source Code
"""

import os
from git import Repo
from .bast import __version__
from .migration import CreateMigration, Migration
import shutil
import click
import re
from subprocess import call
from .environment import load_env
from secrets import token_hex
from colorama import init, Fore, Back

""" Handles the CLI commands and their respective Arguments """


def check():
    server = os.path.abspath('.') + "/server.py"
    config = os.path.abspath('.') + "/config"

    if os.path.exists(server) and os.path.exists(config):
        return True

    return False


@click.group()
@click.version_option(__version__)
def main():
    init(autoreset=True)
    pass


@main.command('create:controller', short_help='Creates a Controller File')
@click.argument('filename', required=1)
def controller_creatr(filename):
    """Name of the controller file to be created"""
    if not check():
        click.echo(Fore.RED + 'ERROR: Ensure you are in a bast app to run the create:controller command')
        return

    path = os.path.abspath('.') + '/controller'
    if not os.path.exists(path):
        os.makedirs(path)

    # if os.path.isfile(path + )

    file_name = str(filename + '.py')
    if os.path.isfile(path+"/" + file_name):
        click.echo(Fore.WHITE + Back.RED + "ERROR: Controller file exists")
        return
    controller_file = open(os.path.abspath('.') + '/controller/' + file_name, 'w+')
    compose = "from bast import Controller\n\nclass " + filename + "(Controller):\n    pass"
    controller_file.write(compose)
    controller_file.close()
    click.echo(Fore.GREEN + "Controller " + filename + " created successfully")


@main.command('create:middleware', short_help="Creates a Middleware")
@click.argument('filename', required=1)
def middleware_creatr(filename):
    if not check():
        click.echo(Fore.RED + 'ERROR: Ensure you are in a bast app to run the create:middleware command')
        return

    path = os.path.abspath('.') + '/middleware'
    if not os.path.exists(path):
        os.makedirs(path)

    file_name = str(filename) + '.py'
    middleware_file = open(os.path.abspath('.') + '/middleware/' + file_name, 'w+')
    compose = "class " + filename + ":\n    def handle(self, request):\n    return True"
    middleware_file.write(compose)
    middleware_file.close()
    click.echo(Fore.GREEN + "Middleware " + filename + " created successfully")


@main.command('create:view', short_help="Create a View File")
@click.argument('filename', required=1)
def view_creatr(filename):
    """Name of the View File to be created"""
    if not check():
        click.echo(Fore.RED + 'ERROR: Ensure you are in a bast app to run the create:view command')
        return

    path = os.path.abspath('.') + '/public/templates'
    if not os.path.exists(path):
        os.makedirs(path)

    filename_ = str(filename + ".html").lower()
    view_file = open(path + "/" + filename_, 'w+')
    view_file.write("")
    view_file.close()
    click.echo(Fore.GREEN + "View file " + filename_ + "created in public/template folder")


@main.command('generate:key', short_help="Generate the APP KEY")
@click.argument('path', required=1)
def make_key(path):
    env_path = os.path.join(path, '.env')
    if not os.path.isfile(env_path):
        click.echo(Fore.RED + ".env file not found. Scaffold a project to generate a key")
        return

    key = token_hex(16)
    with open(env_path, 'r') as file:
        env_data = file.readlines()

    for line_number, line in enumerate(env_data):
        if line.startswith('APP_KEY='):
            env_data[line_number] = 'APP_KEY={0}\n'.format(key)
            break

    with open(env_path, 'w') as file:
        file.writelines(env_data)

    click.echo(Fore.GREEN + "Key Generated successfully: " + key)


@main.command('run', short_help="Run your Bast Server")
@click.option('--serverfile', help="Name of the file to run", default='server.py')
def run(serverfile):
    if not check():
        click.echo(Fore.RED + 'ERROR: Ensure you are in a bast app to use the "run" command')
        return

    call(['python', serverfile])


@main.command('new', short_help="Create a new Bast Project")
@click.argument('projectname', required=1)
def create_new(projectname):
    """Name of the project"""
    git_url = "https://github.com/moluwole/Bast_skeleton"
    path = os.path.abspath('.') + "/" + projectname
    if not os.path.exists(path):
        os.makedirs(path)

    click.echo(Fore.GREEN + '    ___  ___   __________')
    click.echo(Fore.GREEN + '  / _ )/ _ | / __/_  __/')
    click.echo(Fore.GREEN + ' / _  / __ |_\ \  / /')
    click.echo(Fore.GREEN + '/____/_/ |_/___/ /_/')

    click.echo(Fore.GREEN + "Creating Project at %s.... " % path)
    click.echo(Fore.GREEN + "Pulling Project Skeleton from Repo")
    try:
        Repo.clone_from(git_url, path)

        click.echo(Fore.GREEN + "Setting up project")

        shutil.rmtree(path + "/.git")

        if not os.path.exists('/.env'):
            shutil.copy(path + '/.env.example', path + '/.env')

        env_file = path + "/.env"
        if not os.path.isfile(env_file):
            shutil.copy('.env.example', '.env')
        call(['panther', 'generate:key', path])

        click.echo(Fore.GREEN + "New Bast Project created at  %s " % path)
    except Exception as e:
        click.echo(Fore.RED + "An error occurred creating a new project. Try Again.\n Reason: {}".format(e))


@main.command('create:migration', short_help="Create a migration file")
@click.argument('migration_file', required=1)
@click.option('--create', default=True, help="Create the table. OPTIONAL")
@click.option('--table', default=None, help="Name of the table to be created. OPTIONAL")
def migration_creatr(migration_file, create, table):
    """Name of the migration file"""
    if not check():
        click.echo(Fore.RED + 'ERROR: Ensure you are in a bast app to run the create:migration command')
        return

    migration = CreateMigration()
    if table is None:
        table = snake_case(migration_file)
    file = migration.create_file(snake_case(migration_file), table=table, create=create)
    click.echo(Fore.GREEN + 'Migration file created at %s' % file)


@main.command('migration:run', short_help="Run Migration")
@click.option('--pretend', default=False, help="Simulates the Migration")
def migration_run(pretend):
    if not check():
        click.echo(Fore.RED + 'ERROR: Ensure you are in a bast app to run the migration:run command')
        return

    migration = Migration()
    migration.run_(pretend)
    click.echo(Fore.GREEN + 'Migration Run successful')


@main.command('migration:rollback', short_help="Roll Back last Migration")
@click.option('--pretend', default=False, help="Simulates the Migration")
def migration_rollback(pretend):
    if not check():
        click.echo(Fore.RED + 'ERROR: Ensure you are in a bast app to run the migration:rollback command')
        return

    migration = Migration()
    count = migration.rollback_(pretend)
    click.echo(Fore.GREEN + 'Roll Back Executed. %s migrations rolled back' % count)


@main.command('migration:reset', short_help="Reset Migration")
@click.option('--pretend', default=False, help="Simulate the Migration")
def migration_reset(pretend):
    if not check():
        click.echo(Fore.RED + 'ERROR: Ensure you are in a bast app to run the migration:rollback command')
        return

    migration = Migration()
    count = migration.reset_(pretend)
    click.echo(Fore.GREEN + 'Migration Reset successful. %s migrations has been reset' % count)


@main.command('create:model', short_help="Create Model File")
@click.argument('model_file', required=1)
@click.option('--migration', default=True, help="Generate Migration File Also")
def model_creatr(model_file, migration):
    if not check():
        click.echo(Fore.RED + 'ERROR: Ensure you are in a bast app to run the create:model command')
        return

    filename = snake_case(model_file) + ".py"

    directory_path = os.path.abspath('.') + '/models/'
    if not os.path.exists(directory_path):
        os.makedirs(directory_path)

    path = os.path.abspath('.') + '/models/' + filename
    file_open = open(path, 'w+')
    compose = 'from bast import Models\n\nclass %s(Models):\n    __table__ = \'%s\'' \
              % (model_file, snake_case(model_file))
    file_open.write(compose)
    file_open.close()
    if migration:
        migrate = CreateMigration()
        migrate.create_file(name=snake_case(model_file), table=snake_case(model_file), create=True)
    click.echo(Fore.GREEN + '%s has been created at /models' % filename)


def snake_case(string_name):
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', string_name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()



"""
    Bast Web Framework
    (c) Majiyagbe Oluwole <oluwole564@gmail.com>

    For full copyright and license information, view the LICENSE distributed with the Source Code
"""

from tornado.web import HTTPError


class BastException(HTTPError):
    def __init__(self, status_code, message, *args):
        self.message = message
        self.status_code = status_code
        super(BastException, self).__init__(status_code, message, *args)



"""
    Bast Web Framework
    (c) Majiyagbe Oluwole <oluwole564@gmail.com>

    For full copyright and license information, view the LICENSE distributed with the Source Code
"""

from tornado.escape import json_decode, json_encode


class Json:
    """
     Encode and Decode Json values from dictionary objects using Tornado's Json Encode and Decode
     (http://www.tornadoweb.org/en/stable/escape.html)
    """

    @classmethod
    def encode(cls, json):
        """
        Checks the type of data passed to the be encoded.
        :param json:
        :return:
        """
        if type(json) is dict:
            return json_encode(json)
        return json_encode({'message': 'Not a Dictionary object'})

    @classmethod
    def decode(cls, json):
        """
        Checks the type of data to be decoded
        :param json:
        :return:
        """
        if type(json) is dict:
            return json_decode(json)
        return "Not a Dictionary Object"



"""
    Bast Web Framework
    (c) Majiyagbe Oluwole <oluwole564@gmail.com>

    For full copyright and license information, view the LICENSE distributed with the Source Code
"""

import importlib
import logging
import traceback

from tornado.web import RequestHandler
from tornado.util import unicode_type

from .exception import BastException
from .json_ import Json as json_
from .view import TemplateRendering
import os
from tornado.gen import coroutine
from bast import Bast


class Controller(RequestHandler, TemplateRendering):
    method = None
    middleware = None
    providers = {}

    def __init__(self, application, request, **kwargs):
        super().__init__(application, request, **kwargs)
        self.request        = request
        self.application    = application
        self.session_driver = os.getenv("SESSION")

        self.session = Bast.session['session']

    def write_error(self, status_code, **kwargs):
        """
        Handle Exceptions from the server. Formats the HTML into readable form
        """
        reason = self._reason

        if self.settings.get("serve_traceback") and "exc_info" in kwargs:
            error = []
            for line in traceback.format_exception(*kwargs["exc_info"]):
                error.append(line)
        else:
            error = None
        data = {'_traceback': error, 'message': reason, 'code': status_code}
        content = self.render_exception(**data)
        self.write(content)

    def view(self, template_name, kwargs=None):
        """
        Used to render template to view

        Sample usage
        +++++++++++++
        .. code:: python

            from bast import Controller

            class MyController(Controller):
                def index(self):
                    self.view('index.html')
        """
        if kwargs is None:
            kwargs = dict()

        self.add_('session', self.session)

        content = self.render_template(template_name, **kwargs)
        self.write(content)

    def data_received(self, chunk):
        pass

    def __run_middleware__(self, middleware_list):
        """
        Gets the middleware attached to the route and executes it before the route is called. Middlewares in Bast are run before the
        Controller Logic is executed. Returns true once it has been run successfully
        """
        middleware_location = 'middleware'
        return_value = False
        try:
            for func in middleware_list:
                middleware_func = importlib.import_module('{0}.'.format(middleware_location) + func)
                if hasattr(middleware_func, func):
                    class_name = getattr(middleware_func, func)
                    handle = getattr(class_name, 'handle')
                    return_value = handle(class_name, self)
                    return return_value
        except Exception as e:
            print("There is an Error in your Middleware ", e)

        return return_value

    def initialize(self, method, middleware):
        """
        Overridden initialize method from Tornado. Assigns the controller method and middleware attached to the route being executed
        to global variables to be used
        """
        self.method = method
        self.middleware = middleware

    def only(self, arguments):
        """
        returns the key, value pair of the arguments passed as a dict object

        Sample Usage
        ++++++++++++++
        .. code:: python

            from bast import Controller

            class MyController(Controller):
                def index(self):
                    data = self.only(['username'])

        Returns only the argument username and assigns it to the data variable.
        """
        data = {}
        if not isinstance(arguments, list):
            arguments = list(arguments)

        for i in arguments:
            data[i] = self.get_argument(i)
        return data

    def all(self):
        """
        Returns all the arguments passed with the request

        Sample Usage
        ++++++++++++
        .. code:: python

            from bast import Controller

            class MyController(Controller):
                def index(self):
                    data = self.all()

        Returns a dictionary of all the request arguments

        """
        data = {}
        args = self.request.arguments
        for key, value in args.items():
            data[key] = self.get_argument(key)

        return data

    def except_(self, arguments):
        """
        returns the arguments passed to the route except that set by user

        Sample Usage
        ++++++++++++++
        .. code:: python

            from bast import Controller

            class MyController(Controller):
                def index(self):
                    data = self.except_(['arg_name'])

        Returns a dictionary of all arguments except for that provided by as ``arg_name``
        """

        if not isinstance(arguments, list):
            arguments = list(arguments)

        args = self.request.arguments
        data = {}
        for key, value in args.items():
            if key not in arguments:
                data[key] = self.get_argument(key)
        return data

    def json(self, data):
        """
        Encodes the dictionary being passed to JSON and sets the Header to application/json
        """
        self.write(json_.encode(data))
        self.set_header('Content-type', 'application/json')

    @coroutine
    def get(self, *args, **kwargs):
        try:
            if self.middleware is not None and len(self.middleware) > 0:
                value = self.__run_middleware__(self.middleware)
                if not value:
                    return
            func = getattr(self, self.method)
            if func:
                func()
            else:
                raise BastException(404, "Not Found")
        except AttributeError as e:
            logging.error(str(e))
            raise BastException(500, "Controller Function not found")

    @coroutine
    def post(self, *args, **kwargs):
        try:
            if self.middleware is not None and len(self.middleware) > 0:
                value = self.__run_middleware__(self.middleware)
                if not value:
                    return
            func = getattr(self, self.method)
            if func:
                func()
            else:
                raise BastException(404, "Not Found")
        except AttributeError as e:
            logging.error(str(e))
            raise BastException(500, "Controller Function not found")

    @coroutine
    def put(self, *args, **kwargs):
        try:
            if self.middleware is not None and len(self.middleware) > 0:
                value = self.__run_middleware__(self.middleware)
                if not value:
                    return
            func = getattr(self, self.method)
            if func:
                func()
            else:
                raise BastException(404, "Not Found")
        except AttributeError as e:
            logging.error(str(e))
            raise BastException(500, "Controller Function not found")

    @coroutine
    def delete(self, *args, **kwargs):
        try:
            if self.middleware is not None and len(self.middleware) > 0:
                value = self.__run_middleware__(self.middleware)
                if not value:
                    return
            func = getattr(self, self.method)
            if func:
                func()
            else:
                raise BastException(404, "Not Found")
        except AttributeError as e:
            logging.error(str(e))
            raise BastException(500, "Controller Function not found")

    def get_argument(self, name, default=None, strip=True):
        """
        Returns the value of the argument with the given name.

        If default is not provided, returns ``None``

        If the argument appears in the url more than once, we return the last value.

        The returned value is always unicode
        """
        return self._get_argument(name, default, self.request.arguments, strip)[name]

    def headers(self):
        """
        Returns all headers associated with the request
        """
        return self.request.headers

    def header(self, param):
        """
        Returns the header specified by the key provided
        """
        return self.request.headers.get(param)

    def _get_argument(self, name, default, source, strip=True):
        args = self._get_arguments(name, source, strip=strip)
        if not args:
            if default is None:
                return None
        return args[-1]

    def _get_arguments(self, name, source, strip=True):
        values = []
        for v in source.get(name, []):
            v = self.decode_argument(v, name=name)
            if isinstance(v, unicode_type):
                v = self._remove_control_chars_regex.sub(" ", v)
            if strip:
                v = v.strip()

            v = {name: v}
            values.append(v)
        return values



import dotenv
import os
import shutil


def load_env():
    env_path = os.path.abspath('.')

    env_file = env_path + "/.env"
    if not os.path.isfile(env_file):
        shutil.copy('.env.example', '.env')
    return dotenv.load(env_file)




"""
    Bast Web Framework
    (c) Majiyagbe Oluwole <oluwole564@gmail.com>

    For full copyright and license information, view the LICENSE distributed with the Source Code
"""

import sys
import os
from tornado.httpserver import HTTPServer
from tornado.ioloop import IOLoop
from tornado.options import define, options, parse_command_line
from tornado.web import Application, StaticFileHandler
from .environment import load_env
from colorama import init, Fore

__author__ = "Majiyagbe Oluwole"
__copyright__ = ""
__license__ = "M.I.T License"
__version__ = "1.0"
__status__ = "Under Development"


class Bast(Application):
    image_folder    = ""
    script_folder   = ""
    css_folder      = ""
    template_folder = ""

    providers = {}
    session = {}

    def __init__(self, route, **settings):
        """
         Bast Server Class. Runs on Tornado HTTP Server (http://www.tornadoweb.org/en/stable/)

        Constructor for the Bast Server. Takes an instance of the route as parameter.
        The Web handler with routes are handled here.

        Config files are also loaded from the config/config.ini folder.
        Appropriate configurations are loaded from the config file into the os environment for use
        :param route:
        """
        # self.settings = settings
        super().__init__(**settings)
        init()
        self.host = '127.0.0.1'
        self.port = 2000
        self.debug = True

        # self.load_config()
        load_env()
        self.config()

        self.handler = route.all().url
        self.handler.append((r'/css/(.*)', StaticFileHandler, {"path": self.css_folder}))
        self.handler.append((r'/script/(.*)', StaticFileHandler, {"path": self.script_folder}))
        self.handler.append((r'/images/(.*)', StaticFileHandler, {"path": self.image_folder}))

        # append the URL for static files to exception
        self.handler.append((r'/exp/(.*)', StaticFileHandler, {'path': os.path.join(os.path.dirname(os.path.realpath(__file__)), "exception")}))

    def run(self):
        """
        Function to Run the server. Server runs on host: 127.0.0.1 and port: 2000 by default. Debug is also set to false
        by default

        Can be overriden by using the config.ini file
        """
        define("port", default=self.port, help="Run on given port", type=int)
        define("host", default=self.host, help="Run on given host", type=str)
        define("debug", default=self.debug, help="True for development", type=bool)

        parse_command_line()

        print(Fore.GREEN + "Starting Bast Server....")
        print(Fore.GREEN + "Bast Server Running on %s:%s" % (options.host, options.port))

        application = Application(self.handler, debug=options.debug)
        server = HTTPServer(application)
        server.listen(options.port, options.host)
        IOLoop.current().start()

    def config(self):
        sys.path.extend([os.path.abspath('.')])
        from config import config

        static_files    = config.STATIC_FILES
        if config.SESSION_DRIVER is 'memory':
            from bast import MemorySession
            self.session.update({"session": MemorySession()})
        else:
            from bast import FileSession
            self.session.update({'session': FileSession()})

        # providers       = provider.providers

        # print(providers['session'])

        os.environ['TEMPLATE_FOLDER'] = os.path.join(os.path.abspath('.'), static_files['template'])

        self.image_folder   = os.path.join(os.path.abspath('.'), static_files['images'])
        self.css_folder     = os.path.join(os.path.abspath('.'), static_files['css'])
        self.script_folder  = os.path.join(os.path.abspath('.'), static_files['script'])
        # self.providers.update(providers)




from setuptools import setup


def readme():
    with open('README.rst') as f:
        return f.read()


setup(name='Bast',
      version='0.0.1.1.5',
      description='Simple yet Elegant Web Framework with MVC Patterns',
      long_description=readme(),
      classifiers=[
          'Development Status :: 4 - Beta',
          'Environment :: Web Environment',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: MIT License',
          'Natural Language :: English',
          'Operating System :: OS Independent',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.0',
          'Programming Language :: Python :: 3.1',
          'Programming Language :: Python :: 3.2',
          'Programming Language :: Python :: 3.3',
          'Programming Language :: Python :: 3.4',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
          'Topic :: Software Development :: Libraries :: Application Frameworks',
          'Topic :: Software Development :: Libraries :: Python Modules'
      ],
      keywords='web framework tornado ORM MVC',
      url='https://github.com/moluwole/Bast',
      author='Majiyagbe Oluwole',
      author_email='oluwole564@gmail.com',
      maintainer='Majiyagbe Oluwole | Azeez Abiodun Solomon',
      license='MIT',
      packages=['bast'],
      platforms='any',
      install_requires=[
          'tornado',
          'bcrypt',
          'jinja2',
          'gitpython',
          'click',
          'orator',
          'python-env',
          'colorama'
      ],
      entry_points={
          'console_scripts': ['panther=bast.cli:main'],
      },
      include_package_data=True,
      zip_safe=False)



"""
    Bast Web Framework
    (c) Majiyagbe Oluwole <oluwole564@gmail.com>

    For full copyright and license information, view the LICENSE distributed with the Source Code
"""
import importlib


class Route(object):
    """
    Route Class. Appends the URL, Controller Instance and Method to a list instance to be passed on to the server
    instance

    """

    url_ = None
    controller = None
    method = None
    controller_location = 'controller'
    url = []

    # def prefix(self, pref):
    #     return self

    def middleware(self, args):
        """
        Appends a Middleware to the route which is to be executed before the route runs
        """
        if self.url[(len(self.url) - 1)] == (self.url_, self.controller, dict(method=self.method, middleware=None)):
            self.url.pop()
        self.url.append((self.url_, self.controller, dict(method=self.method, middleware=args)))
        return self

    def __return_controller__(self, controller):
        if isinstance(controller, str):
            ctr = controller.split('.')
            if ctr[0].startswith('/'):
                self.controller_location = '.'.join(ctr[0].replace('/', '').split('.')[0:-1])
        else:
            raise Exception

        get_controller = ctr[0].split('.')[-1]
        try:
            # Import the module
            if isinstance(controller, str):
                controller_name = importlib.import_module('{0}.'.format(self.controller_location) + get_controller)
            else:
                controller_name = importlib.import_module('{0}'.format(self.controller_location))

            # Get the controller from the module
            controller_class = getattr(controller_name, get_controller)

            # Get the controller method from the class which in this case is the string
            controller_method = ctr[1]
            return controller_class, controller_method

        except Exception as e:
            print('\033[93mError in your routes/link.py!', e, '\033[0m')

    def get(self, url, controller):
        """
        Gets the Controller and adds the route, controller and method to the url list for GET request
        """
        controller_class, controller_method = self.__return_controller__(controller)

        self.controller = controller_class
        self.method = controller_method
        self.url_ = url

        self.url.append((url, controller_class, dict(method=controller_method, middleware=None)))
        return self

    def post(self, url, controller):
        """
        Gets the Controller and adds the route, controller and method to the url list for the POST request
        """
        controller_class, controller_method = self.__return_controller__(controller)

        self.controller = controller_class
        self.method = controller_method
        self.url_ = url

        self.url.append((url, controller_class, dict(method=controller_method, middleware=None)))
        return self

    def put(self, url, controller):
        """
        Gets the Controller and adds the route, controller and method to the url list for PUT request
        """
        controller_class, controller_method = self.__return_controller__(controller)

        self.controller = controller_class
        self.method = controller_method
        self.url_ = url

        self.url.append((url, controller_class, dict(method=controller_method, middleware=None)))
        return self

    def delete(self, url, controller):
        """
        Gets the Controller and adds the route, controller and method to the url list for the DELETE request
        """
        controller_class, controller_method = self.__return_controller__(controller)

        self.controller = controller_class
        self.method = controller_method
        self.url_ = url

        self.url.append((url, controller_class, dict(method=controller_method, middleware=None)))
        return self

    def all(self):
        """
        Returns the list of URL. Used by Server to get the list of URLS. This is passed to the Bast HTTP Server
        """
        return self



import shelve
import os


class Session(object):
    flash_session = {}
    session = {}

    def __init__(self, session_object):
        self.session.update(session_object)
        self.client_key = os.getenv('APP_KEY')

    def get_data(self):
        session_ = {}
        if self.client_key in self.session:
            session_.update(self.session[self.client_key])

        if self.client_key in self.flash_session:
            session_.update(self.flash_session[self.client_key])

        if not session_:
            return None

        return session_

    def flush(self):
        self.session = {}

    def get(self, key):
        session = self.get_data()
        if session and key in session:
            return session[key]
        return None

    def put(self, key, value):
        if self.client_key not in self.session:
            self.session[self.client_key] = {}

        self.session[self.client_key][key] = value

    def count(self):
        session = self.get_data()
        return len(session)

    def flash(self, key, value):
        if self.client_key not in self.flash_session:
            self.flash_session[self.client_key] = {}

        self.flash_session[self.client_key][key] = value

    def reset(self, flash_only=False):
        if flash_only:
            if self.client_key in self.flash_session:
                self.flash_session[self.client_key] = {}
            else:
                self.session[self.client_key] = {}

    def flush(self, key):
        session = self.get_data()

        if key in session:
            del session[key]
            return True

        return False

    def has(self, key):
        session = self.get_data()
        if session and key in session:
            return True
        return False


class MemorySession(Session):

    def __init__(self):
        SESSION = {}
        Session.__init__(self, SESSION)


class FileSession(Session):
    SHELVE = None

    def __init__(self):
        """

        """
        self.SHELVE = shelve.open('bast.session', writeback=True)
        Session.__init__(self, self.SHELVE)
        # print(self.SHELVE.)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.SHELVE.close()


class RedisSession(Session):
    pass
